# generated by datamodel-codegen:
#   filename:  config-schema.json
#   timestamp: 2024-12-04T08:15:20+00:00

from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import List, Optional

from pydantic import BaseModel, Field

from olot.utils.types import MapStringString, MapStringObject


class Type(Enum):
    layers = 'layers'


class Rootfs(BaseModel):
    diff_ids: List[str]
    type: Type


class HistoryItem(BaseModel):
    created: Optional[datetime] = None
    author: Optional[str] = None
    created_by: Optional[str] = None
    comment: Optional[str] = None
    empty_layer: Optional[bool] = None


# class MapStringObject(BaseModel):
#     __root__: Dict[constr(regex=r'.{1,}'), Dict[str, Any]]


# class Int8(BaseModel):
#     __root__: conint(ge=-128, le=127)


# class Int16(BaseModel):
#     __root__: conint(ge=-32768, le=32767)


# class Int32(BaseModel):
#     __root__: conint(ge=-2147483648, le=2147483647)


# class Int64(BaseModel):
#     __root__: conint(ge=-9223372036854776000, le=9223372036854776000)


# class Uint8(BaseModel):
#     __root__: conint(ge=0, le=255)


# class Uint16(BaseModel):
#     __root__: conint(ge=0, le=65535)


# class Uint32(BaseModel):
#     __root__: conint(ge=0, le=4294967295)


# class Uint64(BaseModel):
#     __root__: conint(ge=0, le=18446744073709552000)


# class Uint16Pointer(BaseModel):
#     __root__: Optional[Uint16]


# class Uint64Pointer(BaseModel):
#     __root__: Optional[Uint64]


# class Base64(BaseModel):
#     __root__: str


# class StringPointer(BaseModel):
#     __root__: Optional[str]


# class MapStringString(BaseModel):
#     __root__: Dict[constr(regex=r'.{1,}'), str]


class Config(BaseModel):
    User: Optional[str] = None
    ExposedPorts: Optional[MapStringObject] = None
    Env: Optional[List[str]] = None
    Entrypoint: Optional[List[str]] = None
    Cmd: Optional[List[str]] = None
    Volumes: Optional[MapStringObject] = None
    WorkingDir: Optional[str] = None
    Labels: Optional[MapStringString] = None
    StopSignal: Optional[str] = None
    ArgsEscaped: Optional[bool] = None


class OCIManifestConfig(BaseModel):
    created: Optional[datetime] = None
    author: Optional[str] = None
    architecture: str
    variant: Optional[str] = None
    os: str
    os_version: Optional[str] = Field(None, alias='os.version')
    os_features: Optional[List[str]] = Field(None, alias='os.features')
    config: Optional[Config] = None
    rootfs: Rootfs
    history: Optional[List[HistoryItem]] = None



from pathlib import Path
import os
import json
import argparse
from typing import List

from olot.oci.oci_image_manifest import create_oci_image_manifest, create_manifest_layers
from olot.oci.oci_image_layout import create_ocilayout
from olot.oci.oci_common import MediaTypes, Values
from olot.oci.oci_image_index import Manifest, create_oci_image_index
from olot.utils.files import MIMETypes, tarball_from_file, targz_from_file
from olot.utils.types import compute_hash_of_str

def create_oci_artifact_from_model(source_dir: Path, dest_dir: Path):
    """
    Create an OCI artifact from a model directory.

    Args:
        source_dir: The directory containing the model files.
        dest_dir: The directory to write the OCI artifact to. If None, a directory named 'oci' will be created in the source directory.
    """
    if not source_dir.exists():
        raise NotADirectoryError(f"Input directory '{source_dir}' does not exist.")

    if dest_dir is None:
        dest_dir = source_dir / "oci"
    os.makedirs(dest_dir, exist_ok=True)

    sha256_path = dest_dir / "blobs" / "sha256"
    os.makedirs(sha256_path, exist_ok=True)

    # assume flat structure for source_dir for now
    # TODO: handle subdirectories appropriately
    model_files = [source_dir / Path(f) for f in os.listdir(source_dir) if os.path.isfile(os.path.join(source_dir, f))]

    # Populate blobs directory
    layers = create_blobs(model_files, dest_dir)

    # Create the OCI image manifest
    manifest_layers = create_manifest_layers(model_files, layers)
    artifactType = MIMETypes.mlmodel
    manifest = create_oci_image_manifest(
        artifactType=artifactType,
        layers=manifest_layers,
    )
    manifest_json = json.dumps(manifest.dict(exclude_none=True), indent=4, sort_keys=True)
    manifest_SHA = compute_hash_of_str(manifest_json)
    with open(sha256_path / manifest_SHA, "w") as f:
        f.write(manifest_json)

    # Create the OCI image index
    index = create_oci_image_index(
        manifests = [
            Manifest(
                mediaType=MediaTypes.manifest,
                size=os.stat(sha256_path / manifest_SHA).st_size,
                digest=f"sha256:{manifest_SHA}",
                urls = None,
            )
        ]
    )
    index_json = json.dumps(index.dict(exclude_none=True), indent=4, sort_keys=True)
    with open(dest_dir / "index.json", "w") as f:
        f.write(index_json)

    # Create the OCI-layout file
    oci_layout = create_ocilayout()
    with open(dest_dir / "oci-layout", "w") as f:
        f.write(json.dumps(oci_layout.model_dump(), indent=4, sort_keys=True))

    # Create empty config file with digest as name
    empty_config: dict[str, str] = {}
    empty_digest_split = Values.empty_digest.split(":")
    if len(empty_digest_split) == 2:
        with open(dest_dir / "blobs" / "sha256" / empty_digest_split[1], "w") as f:
            f.write(json.dumps(empty_config))
    else:
        raise ValueError(f"Invalid empty_digest format: {Values.empty_digest}")

def create_blobs(model_files: List[Path], dest_dir: Path):
    """
    Create the blobs directory for an OCI artifact.
    """
    layers = {} # layer digest : (precomp, postcomp)
    sha256_path = dest_dir / "blobs" / "sha256"

    for model_file in model_files:
        file_name = os.path.basename(os.path.normpath(model_file))
        # handle model card file if encountered - assume README.md is the modelcard
        if file_name.endswith("README.md"):
            postcomp_chksum, precomp_chksum = targz_from_file(model_file, sha256_path)
            layers[file_name] = (precomp_chksum, postcomp_chksum)
        else:
            checksum = tarball_from_file(model_file, sha256_path)
            layers[file_name] = (checksum, "")
    return layers

# create a main function to test the function
def main():
    parser = argparse.ArgumentParser(description="Create OCI artifact from model")
    parser.add_argument('source_dir', type=str, help='Path to the source directory')
    args = parser.parse_args()

    source_dir = Path(args.source_dir)
    create_oci_artifact_from_model(source_dir, None)

if __name__ == "__main__":
    main()



import os
import shutil
import subprocess
import typing

def is_skopeo() -> bool :
    return shutil.which("skopeo") is not None


def skopeo_pull(base_image: str, dest: typing.Union[str, os.PathLike]):
    if isinstance(dest, os.PathLike):
        dest = str(dest)
    return subprocess.run(["skopeo", "copy", "--multi-arch", "all", "--remove-signatures", "docker://"+base_image, "oci:"+dest+":latest"], check=True)


def skopeo_push(src: typing.Union[str, os.PathLike], oci_ref: str):
    if isinstance(src, os.PathLike):
        src = str(src)
    return subprocess.run(["skopeo", "copy", "--multi-arch", "all", "--dest-tls-verify=false", "oci:"+src+":latest", "docker://"+oci_ref], check=True)



# generated by datamodel-codegen:
#   filename:  image-layout-schema.json
#   timestamp: 2024-12-04T11:33:36+00:00

from __future__ import annotations

from enum import Enum
from pathlib import Path

from pydantic import BaseModel, Field


class ImageLayoutVersion(Enum):
    field_1_0_0 = '1.0.0'


class OCIImageLayout(BaseModel):
    imageLayoutVersion: ImageLayoutVersion = Field(
        ..., description='version of the OCI Image Layout (in the oci-layout file)'
    )

    class Config:
        use_enum_values = True

def verify_ocilayout(ocilayout: Path):
    with open(ocilayout / "oci-layout", "r") as f:
        m = OCIImageLayout.model_validate_json(f.read())
        if not m.imageLayoutVersion == ImageLayoutVersion.field_1_0_0.value:
            raise ValueError(f"Unexpected ocilayout in {ocilayout}")
        else:
            return True

def create_ocilayout() -> OCIImageLayout:
    return OCIImageLayout(imageLayoutVersion=ImageLayoutVersion.field_1_0_0)


# generated by datamodel-codegen:
#   filename:  defs-descriptor.json
#   timestamp: 2024-12-05T11:15:48+00:00

from __future__ import annotations

from pydantic import BaseModel, Field


class MediaType(BaseModel):
    id: str
    type: str
    pattern: str


class Digest(BaseModel):
    description: str
    type: str
    pattern: str


class Items(BaseModel):
    type: str
    format: str


class Urls(BaseModel):
    description: str
    type: str
    items: Items


class Annotations(BaseModel):
    field_ref: str = Field(..., alias='$ref')


class Definitions(BaseModel):
    mediaType: MediaType
    digest: Digest
    urls: Urls
    annotations: Annotations


class Model(BaseModel):
    description: str
    definitions: Definitions



from typing import Annotated, Any, Dict
from pydantic import Field
import hashlib

NonEmptyString = Annotated[str, Field(..., pattern=r".{1,}")]
MapStringString = Annotated[Dict[NonEmptyString, str], Field(...)]
MapStringObject = Annotated[Dict[NonEmptyString, Any], Field(...)]

Int8 = Annotated[int, Field(ge=-128, le=127)]
Int64 = Annotated[int, Field(ge=-9223372036854776000, le=9223372036854776000)]
Base64 = Annotated[str, Field()]
Annotations = Annotated[MapStringString, Field()]

def compute_hash_of_str(content: str) -> str:
    h = hashlib.sha256()
    h.update(content.encode())
    return h.hexdigest()




import logging
import os
from pathlib import Path
from pprint import pprint
import tarfile
from typing import Dict, List, Sequence
import typing
import click

from olot.oci.oci_config import OCIManifestConfig

from olot.oci.oci_image_index import OCIImageIndex, read_ocilayout_root_index
from olot.oci.oci_image_manifest import OCIImageManifest, ContentDescriptor
from olot.oci.oci_image_layout import verify_ocilayout
from olot.oci.oci_common import MediaTypes

from olot.utils.files import handle_remove, tarball_from_file, targz_from_file
from olot.utils.types import compute_hash_of_str

logger = logging.getLogger(__name__)

def oci_layers_on_top(
        ocilayout: typing.Union[str, os.PathLike],
        model_files: Sequence[os.PathLike],
        modelcard: typing.Union[os.PathLike, None] = None,
        *,
        remove_originals: bool = False):
    """
    Add contents to an oci-layout directory as new blob layers

    Args:
        ocilayout: The oci-layout directory of the base image.
        model_files: PathLike array to be added as new blob layers.
        modelcard: PathLike of the README.md of the ModelCarD, will be added as the last layer with compression and annotations.
        remove: whether to remove the original content files after having added the layers, default: False.
    """
    if not isinstance(ocilayout, Path):
        ocilayout = Path(ocilayout)
    if remove_originals:
        logger.info("Invoked with 'remove' to delete original contents after adding as a blob layer.")

    verify_ocilayout(ocilayout)
    ocilayout_root_index = read_ocilayout_root_index(ocilayout)
    ocilayout_indexes: Dict[str, OCIImageIndex] = crawl_ocilayout_indexes(ocilayout, ocilayout_root_index)
    ocilayout_manifests: Dict[str, OCIImageManifest] = crawl_ocilayout_manifests(ocilayout, ocilayout_indexes)
    new_layers = {} # layer digest : diff_id

    sha256_path = ocilayout / "blobs" / "sha256"
    for model in model_files:
        model = Path(model)
        new_layer = tarball_from_file(model, sha256_path)
        new_layers[new_layer] = new_layer
        if remove_originals:
            handle_remove(model)
    if modelcard is not None:
        modelcard_layer_diffid = targz_from_file(Path(modelcard), sha256_path)
        new_layers[modelcard_layer_diffid[0]] = modelcard_layer_diffid[1]
        if remove_originals:
            handle_remove(modelcard)

    new_ocilayout_manifests: Dict[str, str] = {}
    for manifest_hash, manifest in ocilayout_manifests.items():
        print(manifest_hash, manifest.mediaType)
        config_sha = manifest.config.digest.removeprefix("sha256:")
        mc = None
        with open(ocilayout / "blobs" / "sha256" / config_sha, "r") as cf:
            mc = OCIManifestConfig.model_validate_json(cf.read())
        for layer, diffid in new_layers.items():
            size = os.stat(ocilayout / "blobs" / "sha256" / layer).st_size
            mt = MediaTypes.layer if layer == diffid else MediaTypes.layer_gzip
            la = None if layer == diffid else {"io.opendatahub.temp.layer.type":"modelcard"}
            cd = ContentDescriptor(
                mediaType=mt,
                digest="sha256:"+layer,
                size=size,
                urls=None,
                data=None,
                artifactType=None,
                annotations=la
            )
            mc.rootfs.diff_ids.append("sha256:"+diffid)
            manifest.layers.append(cd)
        # TODO: add to Manifest.config the history/author of this project.
        mc_json = mc.model_dump_json(exclude_none=True)
        with open(ocilayout / "blobs" / "sha256" / config_sha, "w") as cf:
            cf.write(mc_json)
        mc_json_hash = compute_hash_of_str(mc_json)
        os.rename(ocilayout / "blobs" / "sha256" / config_sha, ocilayout / "blobs" / "sha256" / mc_json_hash)
        print(f"Renamed config from: {config_sha} to {mc_json_hash}")
        config_sha = mc_json_hash
        manifest.config.digest = "sha256:" + config_sha
        manifest.config.size = os.stat(ocilayout / "blobs" / "sha256" / config_sha).st_size
        if manifest.annotations is None:
            manifest.annotations = {}
        manifest.annotations["io.opendatahub.temp.author"] = "olot"
        manifest_json = manifest.model_dump_json(exclude_none=True)
        with open(ocilayout / "blobs" / "sha256" / manifest_hash, "w") as cf:
            cf.write(manifest_json)
        manifest_json_hash = compute_hash_of_str(manifest_json)
        os.rename(ocilayout / "blobs" / "sha256" / manifest_hash, ocilayout / "blobs" / "sha256" / manifest_json_hash)
        print(f"Renamed Manifest from: {manifest_hash} to {manifest_json_hash}")
        new_ocilayout_manifests[manifest_hash] = manifest_json_hash
        manifest_hash = manifest_json_hash
    pprint(new_ocilayout_manifests)
    new_ocilayout_indexes: Dict[str, str] = {}
    for index_hash, index in ocilayout_indexes.items():
        print(index_hash, index.mediaType)
        for m in index.manifests:
            lookup_new_hash = new_ocilayout_manifests[m.digest.removeprefix("sha256:")]
            print(f"old manifest {m.digest} is now at {lookup_new_hash}")
            m.digest = "sha256:" + lookup_new_hash
            m.size = os.stat(ocilayout / "blobs" / "sha256" / lookup_new_hash).st_size
        index_json = index.model_dump_json(exclude_none=True)
        with open(ocilayout / "blobs" / "sha256" / index_hash, "w") as idxf:
            idxf.write(index_json)
        index_json_hash = compute_hash_of_str(index_json)
        os.rename(ocilayout / "blobs" / "sha256" / index_hash, ocilayout / "blobs" / "sha256" / index_json_hash)
        print(f"Renamed Index from: {index_hash} to {index_json_hash}")
        new_ocilayout_indexes[index_hash] = index_json_hash
    pprint(new_ocilayout_indexes)
    for entry in ocilayout_root_index.manifests:
        if entry.mediaType == MediaTypes.index:
            lookup_new_hash = new_ocilayout_indexes[entry.digest.removeprefix("sha256:")]
            print(f"old index {entry.digest} is now at {lookup_new_hash}")
            entry.digest = "sha256:" + lookup_new_hash
            entry.size = os.stat(ocilayout / "blobs" / "sha256" / lookup_new_hash).st_size
        elif entry.mediaType == MediaTypes.manifest:
            lookup_new_hash = new_ocilayout_manifests[entry.digest.removeprefix("sha256:")]
            print(f"old index {entry.digest} is now at {lookup_new_hash}")
            entry.digest = "sha256:" + lookup_new_hash
            entry.size = os.stat(ocilayout / "blobs" / "sha256" / lookup_new_hash).st_size
        else:
            raise ValueError(f"unknown root index mediaType {entry.mediaType}")
    with open(ocilayout / "index.json", "w") as root_idx_f:
        root_idx_f.write(ocilayout_root_index.model_dump_json(exclude_none=True))



def crawl_ocilayout_manifests(ocilayout: Path, ocilayout_indexes: Dict[str, OCIImageIndex]) -> Dict[str, OCIImageManifest]:
    ocilayout_manifests: Dict[str, OCIImageManifest]  = {}
    for _, mi in ocilayout_indexes.items():
        for m in mi.manifests:
            print(m)
            if m.mediaType != MediaTypes.manifest:
                raise ValueError("Did not expect something else than Image Manifest in a Index")
            target_hash = m.digest.removeprefix("sha256:")
            print(target_hash)
            manifest_path = ocilayout / "blobs" / "sha256" / target_hash
            with open(manifest_path, "r") as ip:
                ocilayout_manifests[target_hash] = OCIImageManifest.model_validate_json(ip.read())
    return ocilayout_manifests


def crawl_ocilayout_indexes(ocilayout: Path, ocilayout_root_index: OCIImageIndex) -> Dict[str, OCIImageIndex] :
    ocilayout_indexes: Dict[str, OCIImageIndex] = {}
    for m in ocilayout_root_index.manifests:
        if m.mediaType == MediaTypes.index:
            target_hash = m.digest.removeprefix("sha256:")
            index_path = ocilayout / "blobs" / "sha256" / target_hash
            with open(index_path, "r") as ip:
                ocilayout_indexes[target_hash] = OCIImageIndex.model_validate_json(ip.read())
        else:
            if len(ocilayout_indexes) == 0:
                raise ValueError("TODO the root index has Image manifest")
            else:
                click.echo(f"Found Image Manifest {m.digest} in root index, TODO assuming these are referred through the other indexes")
    return ocilayout_indexes


def crawl_ocilayout_blobs_to_extract(ocilayout: Path, 
                                     output_path: Path,
                                     tar_filter_dir: str = "/models") -> List[str]:
    """
    Extract from OCI Image/ModelCar only the contents from a specific directory.

    Args:
        ocilayout: The directory containing the oci-layout of the OCI Image/ModelCar.
        output_path: The directory where to extract the ML model assets from the ModelCar to.
        tar_filter_dir: The subdirectory in the ModelCar to extract, defaults to `"/models"`.

    Returns:
        The list of extracted ML contents from the OCI Image/ModelCar.
    """
    extracted: List[str] = []
    tar_filter_dir= tar_filter_dir.lstrip("/")
    blobs_path = ocilayout / "blobs" / "sha256"
    if not os.path.exists(output_path):
        os.makedirs(output_path)
    verify_ocilayout(ocilayout)
    ocilayout_root_index = read_ocilayout_root_index(ocilayout)
    if len(ocilayout_root_index.manifests) != 1:
        raise ValueError("TODO the root index has more than one manifest, expected single ModelCar")
    manifest0 = ocilayout_root_index.manifests[0]
    if manifest0.mediaType != MediaTypes.manifest:
        raise ValueError("Can only extract from ModelCar Image manifests")
    target_hash = manifest0.digest.removeprefix("sha256:")
    manifest_path = blobs_path / target_hash
    with open(manifest_path, "r") as ip:
        image_manifest = OCIImageManifest.model_validate_json(ip.read())
    for layer in image_manifest.layers:
        if layer.mediaType == MediaTypes.layer or layer.mediaType == MediaTypes.layer_gzip:
            target_hash = layer.digest.removeprefix("sha256:")
            manifest_path = blobs_path / target_hash
            with tarfile.open(manifest_path, "r:*") as tar:
                for member in tar.getmembers():
                    if member.isfile() and member.name.startswith(tar_filter_dir):
                        tar.extract(member, path=output_path)
                        extracted.append(member.name)
    return extracted


if __name__ == "__main__":
    print("?")



from os import PathLike
import click
import logging

from .basics import oci_layers_on_top


@click.command()
@click.option("-m", "--modelcard", type=click.Path(exists=True, file_okay=True, dir_okay=False))
@click.argument('ocilayout', type=click.Path(exists=True, file_okay=False, dir_okay=True))
@click.argument('model_files', nargs=-1)
@click.option('-r', '--remove-originals', is_flag=True)
def cli(ocilayout: str, modelcard: PathLike, model_files, remove_originals: bool):
    logging.basicConfig(level=logging.INFO)
    oci_layers_on_top(ocilayout, model_files, modelcard, remove_originals=remove_originals)
